#include <stdio.h>
#include <stdint.h>
#include <genvector/gds_char.h>
#include <genvector/vtp0.h>
#include <genlist/gendlist.h>

#include "concurrent.h"

extern char *vers;
extern size_t msg_size_max, write_size_max;
extern long rlist_size_max;
extern FILE *logfile;
extern int foreground;

#define logprintf(x...) \
do { \
	if (logfile != NULL) {\
		fprintf(logfile, x); \
		fflush(logfile); \
	} \
} while(0)

#define logprintf_(x...) \
do { \
	if (logfile != NULL) \
		fprintf(logfile, x); \
} while(0)

#define ETODO

typedef struct channel_s channel_t;

typedef struct message_s {
	uint32_t addr;
	int refco;
	mutex_t refco_lock;

	enum {
		MT_MSG,     /* normal message, should be prefixed with {addr} tag during copy to rbuf on addressed channels */
		MT_CTRL     /* control message, no address prefix added on rbuf copy (the {} prefix is already added, e.g. {123:o}) */
	} type;
	channel_t *chan;
	gdl_elem_t link;    /* in list of messages in a channel */

	/* payload */
	int len;
	char msg[1];        /* over-allocated */
} message_t;

typedef struct client_s { /* a client is an active fd invoked with open() on a channel */
	uint32_t id;
	unsigned rd:1;      /* open for read (messages should be relayed here) */
	unsigned wr:1;      /* open for write (messages should be accepted from here) */
	unsigned master:1;
	unsigned dead:1;    /* set to 1 after a release() is received */
	unsigned freeing:1; /* set to 1 when client is being free'd so that it is not free'd twice */
	unsigned rbuf_eof:1;/* stop reading after current rbuf runs out */
	channel_t *chan;

	/* buffers */
	long rlist_waiters;     /* other clients waiting for rlist; do not free client while this is not 0 */
	gds_t rbuf, wbuf;
	vtp0_t rlist;           /* of (message_t *) */
	mutex_t rlist_lock;     /* only one thread can read or write rlist */
	notify_t rlist_notify;  /* poke client if rlist changes from 0 length to non-0 length */

	/* poll */
	void *poll_handle;      /* opaque data used by fuse to track this poll */
	unsigned *poll_revents; /* opaque data used by fuse to track this poll */
	mutex_t poll_lock;      /* protect p[oll fields */

	gdl_elem_t link;    /* in list of clients in a channel */
} client_t;

struct channel_s {
	char *path;
	uint32_t id;
	uint32_t next_client_id;
	char type;           /* one of U, A, S or B */

	/* mutex-locked fields */
	unsigned zombie:1;   /* master left, got unlinked, but some slaves still have pending read */
	client_t *master;    /* NULL or one of the clients in case of type U or S */
	gdl_list_t clients;  /* of client_t */
	gdl_list_t messages; /* of message_t */
	mutex_t lock;        /* only one thread can manipulate the above fields */

	gdl_elem_t link;     /* in channels or zombie_channels */
};

extern gdl_list_t channels;        /* list of all existing channels */
extern gdl_list_t zombie_channels; /* list of all channels that lost master and got unlinked */
extern uint32_t next_channel_id;
extern mutex_t channels_lock;      /* only one thread should manipulate channels and zombie_channels at a time */

#define CHANTYPE_NEEDS_MASTER(type) (((type) == 'U') || ((type) == 'S'))
#define CHANTYPE_AUTOREMOVE(type)   (((type) == 'U') || ((type) == 'S'))

#define chan_client_to_fd(chanID, clientID) ((uint64_t)(chanID) << 32) | ((uint64_t)(clientID))
#define fd_to_chan_client(fd, chanID, clientID) \
do { \
	chanID = fd >> 32; \
	clientID = fd & 0xffffffffULL; \
} while(0)

extern channel_t *statch;

static inline channel_t *id2chan(uint32_t chanID)
{
	channel_t *c;

	if ((statch != NULL) && (statch->id == chanID))
		return statch;

	for(c = gdl_first(&channels); c != NULL; c = gdl_next(&channels, c))
		if (c->id == chanID)
			return c;

	for(c = gdl_first(&zombie_channels); c != NULL; c = gdl_next(&zombie_channels, c))
		if (c->id == chanID)
			return c;

	return NULL;
}

static inline client_t *id2client(channel_t *chan, uint32_t clientID)
{
	client_t *c;

	if (chan == NULL)
		return NULL;

	for(c = gdl_first(&chan->clients); c != NULL; c = gdl_next(&chan->clients, c))
		if (c->id == clientID)
			return c;

	return NULL;
}

/*** Poll callbacks ***/
/* dispatch and other parts of the generic code can call these to inform the
    fuse glue layer (futebus.c) on poll related events per client */

/* called when client is closed/free'd */
void futebus_poll_uninit(client_t *client);

/* called when client becomes readable (anything inserted in the rlist) */
void futebus_poll_readable(client_t *client);

